import math

from pyplanet.views.generics.widget import TimesWidgetView
from pyplanet.utils import times
import logging


class CPWidgetView(TimesWidgetView):
    widget_x = -160
    widget_y = 70.5
    z_index = 130
    size_x = 38
    size_y = 55.5
    title = 'Current CPs'

    template_name = 'currentcps/cpwidget.xml'

    def __init__(self, app, height=55.5):
        super().__init__(self)
        self.app = app
        self.manager = app.context.ui
        self.id = 'pyplanet__widgets_currentcps'

        self.size_y = height

    def set_num_cps(self, num_cps=None):
        if not num_cps:
            self.title = 'Current CPs'
        else:
            self.title = 'Current CPs $aaa/ {}'.format(num_cps)

    async def get_player_data(self):
        data = await super().get_player_data()

        # Calculate the maximum number of rows that can be displayed
        max_n = math.floor((self.size_y - 5.5) / 3.3)

        # Maps a logon to the data that should be displayed
        cps = {}

        for idx, player in enumerate(self.app.instance.player_manager.online):
            last_fin = 0
            list_times = []
            n = 1
            for pcp in self.app.player_cps:
                # Make sure to only display a certain number of entries
                if float(n) >= max_n:
                    break

                # Set time color to green for your own CP time
                list_time = {'index': n, 'color': "$0f3" if player.login == pcp.player.login else "$bbb"}

                # Display 'fin' when the player crossed the finish line else display the CP number
                if pcp.cp == -1 or (pcp.cp == 0 and pcp.time != 0):
                    list_time['cp'] = 'fin'
                    last_fin += 1
                else:
                    list_time['cp'] = str(pcp.cp)

                list_time['cptime'] = times.format_time(pcp.time)
                list_time['nickname'] = pcp.player.nickname
                list_time['login'] = pcp.player.login

                # Only show top 5 fins but always show the current player
                if (pcp.cp == -1 or (pcp.cp == 0 and pcp.time != 0)) and last_fin > 5:
                    if player.login != pcp.player.login:
                        continue
                    list_times[4] = list_time
                    continue

                list_times.append(list_time)
                n += 1
            cps[player.login] = {'cps': list_times}

        data.update(cps)

        return data

    async def get_context_data(self):
        self.widget_y = 12.5 if self.app.dedimania_enabled else 70.5
        return await super().get_context_data()

    async def handle_catch_all(self, player, action, values, **kwargs):
        logging.debug("CatchAll: " + player.login + ": " + action)
        if str(action).startswith('spec_'):
            target = action[5:]
            await self.app.spec_player(player=player, target_login=target)



from pyplanet.contrib.command import Command
from pyplanet.apps.contrib.admin.views.mapbrowser import BrowserView


class MapBrowser:
	def __init__(self, app):
		self.app = app
		self.instance = app.instance

	async def on_start(self):
		await self.instance.permission_manager.register(
			'localmaps', 'Shows a file browser to browse the local files',
			app=self.app, min_level=3
		)

		await self.app.instance.command_manager.register(
			Command(command='localmaps', target=self.show_browser, perms='admin:localmaps', admin=True,
					description='Shows a file browser to browse local files.')
		)

	async def show_browser(self, player, data, **kwargs):
		view = BrowserView(self.app, player)
		await view.set_dir('')



from pyplanet.conf import default_settings

_GLOBAL_DEFAULT = object()


class ConfigBackend:
	"""
	The base config backend class contains everything in place for one-time loading and memory pooled configuration.
	You only have to override the name property and the ``load()`` method to have a file reader working.

	Please make sure you convert all keys to UPPER in your ``load()`` before you store it in ``self.settings``!

	"""
	name = None

	def __init__(self, **options):
		self.options = options
		self.defaults = dict()
		self.settings = dict()
		self.settings['SETTINGS_METHOD'] = self.name

	def load(self):
		"""
		Load method. Override this method and call the ``super().load()`` to load defaults before you load your configuration
		file or backend.
		"""
		# Add default configuration to our cached/static context.
		for setting in dir(default_settings):
			if setting.isupper():
				self.defaults[setting] = getattr(default_settings, setting)

	def is_overriden(self, setting):
		"""
		Is the setting overriden by local settings module?

		:param setting: setting to check
		:type setting: str
		:return: boolean
		:rtype: bool
		"""
		return setting.upper() in self.settings

	def get(self, key, default=_GLOBAL_DEFAULT):
		"""
		Get a setting by key, return the value or if not exists the default parameter given or if the default parameter
		is not given the provided default in the PyPlanet version.

		:param key: Key of setting, case sensitive! All keys are in uppercase!
		:param default: Default override. Leave empty/don't provide to use the defaults by PyPlanet.
		:type key: str
		:return: Value of setting.
		"""
		if key in self.settings:
			return self.settings[key]
		if default != _GLOBAL_DEFAULT:
			return default
		if key in self.defaults:
			return self.defaults[key]
		raise KeyError(key)

	def set(self, key, value):
		"""
		Set a setting by key and set the provided value. Please keep in mind that not all configuration backends
		support setting values!

		:param key: Key of setting, case sensitive! All keys are in uppercase!
		:param value: Value to set.
		:raise: NotImplementedError
		"""
		raise NotImplementedError('Settings can not be written with this backend!')



from pyplanet.views import TemplateView
from pyplanet.views.generics.widget import WidgetView


class SectorTimesWidget(WidgetView):
	# widget_x = 20
	widget_y = -70

	template_name = 'sector_times/sector_times.xml'

	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.sector_times.SectorTimes
		"""
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_sector_times'

	@property
	def widget_x(self):
		if self.app:
			if self.app.instance.game.game == 'tmnext':
				return -55
		return 20

	async def get_per_player_data(self, login):
		dedi_score = 0
		dedi_record = None
		local_score = 0
		local_record = None
		if 'dedimania' in self.app.instance.apps.apps:
			try:
				dedi_record = [x for x in self.app.instance.apps.apps['dedimania'].current_records if x.login == login]
				if len(dedi_record) > 0:
					dedi_record = dedi_record[0]
			except:
				pass
		if 'local_records' in self.app.instance.apps.apps:
			try:
				local_record = [x for x in self.app.instance.apps.apps['local_records'].current_records if x.player.login == login]
				if len(local_record) > 0:
					local_record = local_record[0]
			except:
				pass

		if dedi_record and hasattr(dedi_record, 'score'):
			dedi_score = dedi_record.score
		if local_record and hasattr(local_record, 'score'):
			local_score = local_record.score

		# Get fastest score, source and checkpoint scores.
		fastest_score = 0
		fastest_source = ''
		fastest_cps = list()
		if dedi_score > 0 and (local_score <= 0 or dedi_score < local_score):
			fastest_score = dedi_score
			fastest_cps = dedi_record.cps
			fastest_source = 'Dedi'
		elif local_score > 0 and (dedi_score <= 0 or local_score <= dedi_score):
			fastest_score = local_score
			fastest_cps = local_record.checkpoints
			fastest_source = 'Local'

		if isinstance(fastest_cps, list):
			fastest_cps = ','.join([str(c) for c in fastest_cps])

		context = await super().get_per_player_data(login)
		context['record_sector_times'] = fastest_cps
		context['record_time'] = fastest_score
		context['record_source'] = fastest_source
		if not fastest_cps:
			context['record_time'] = 0

		return context


class CheckpointDiffWidget(WidgetView):
	widget_x = 0
	widget_y = 0

	template_name = 'sector_times/cp_diff.xml'

	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.sector_times.SectorTimes
		"""
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_cp_diff'

	async def get_per_player_data(self, login):
		dedi_score = 0
		dedi_record = None
		local_score = 0
		local_record = None
		if 'dedimania' in self.app.instance.apps.apps:
			try:
				dedi_record = [x for x in self.app.instance.apps.apps['dedimania'].current_records if x.login == login]
				if len(dedi_record) > 0:
					dedi_record = dedi_record[0]
			except:
				pass
		if 'local_records' in self.app.instance.apps.apps:
			try:
				local_record = [x for x in self.app.instance.apps.apps['local_records'].current_records if x.player.login == login]
				if len(local_record) > 0:
					local_record = local_record[0]
			except:
				pass

		if dedi_record and hasattr(dedi_record, 'score'):
			dedi_score = dedi_record.score
		if local_record and hasattr(local_record, 'score'):
			local_score = local_record.score

		# Get fastest score, source and checkpoint scores.
		fastest_score = 0
		fastest_source = 'PB'
		fastest_cps = list()
		if dedi_score > 0 and (local_score <= 0 or dedi_score < local_score):
			fastest_score = dedi_score
			fastest_cps = dedi_record.cps
		elif local_score > 0 and (dedi_score <= 0 or local_score <= dedi_score):
			fastest_score = local_score
			fastest_cps = local_record.checkpoints

		if isinstance(fastest_cps, list):
			fastest_cps = ','.join([str(c) for c in fastest_cps])

		context = await super().get_per_player_data(login)
		context['record_sector_times'] = fastest_cps
		context['record_time'] = fastest_score
		context['record_source'] = fastest_source
		if not fastest_cps:
			context['record_time'] = 0

		return context


class GearIndicatorView(TemplateView):
	template_name = 'sector_times/gear_indicator.xml'

	def __init__(self, app, *args, **kwargs):
		super().__init__(app.context.ui, *args, **kwargs)
		self.app = app
		self.manager = app.context.ui
		self.id = 'sector_times_gear'



"""
Settings for the PyPlanet application.

Values will dynamically read from the settings module provided by your setup. All default values will be overwritten.
Default values are retrieved from `default_settings.py`.
"""
import logging
import os

from pyplanet.conf import default_settings
from pyplanet.conf.backends.json import JsonConfigBackend
from pyplanet.conf.backends.python import PythonConfigBackend
from pyplanet.conf.backends.yaml import YamlConfigBackend
from pyplanet.core.exceptions import ImproperlyConfigured


class LazySettings:
	"""
	THe lazy settings class will cache and merge settings that the settings module provided and the defaults
	of the system.
	"""
	BACKEND = {
		'python': PythonConfigBackend,
		'json': JsonConfigBackend,
		'yaml': YamlConfigBackend,
	}

	def __init__(self):
		self._settings = None

	def _setup(self, optional_loading=False):
		"""
		Setup will create the wrapped settings and load the settings module.
		"""
		settings_method = os.environ.get('PYPLANET_SETTINGS_METHOD', 'python')
		if not settings_method and not optional_loading:
			raise ImproperlyConfigured(
				'Settings method is not defined! Please define PYPLANET_SETTINGS_METHOD in your '
				'environment or start script. The possible values: \'python\', \'json\', \'yaml\''
			)

		self._settings_method = settings_method
		try:
			self._settings_method_class = self.BACKEND[settings_method]
		except:
			raise ImproperlyConfigured(
				'The provided settings method \'{}\' does not exist in the current PyPlanet version!'
				'The current possible methods: \'python\', \'json\', \'yaml\'.'
			)

		# Initiate the backend class.
		self._settings = self._settings_method_class()

		# Load the contents.
		try:
			self._settings.load()
		except Exception as e:
			if not optional_loading:
				logging.exception(e)
				exit(1)

	def __getattr__(self, item):
		"""
		Get value from local or wrapped settings.
		"""
		if self._settings is None:
			self._setup()
		val = self._settings.get(item)
		self.__dict__[item] = val
		return val

	def __setattr__(self, key, value):
		"""
		Set and clear cached settings.
		"""
		if key == '_settings':
			self.__dict__.clear()
		else:
			self.__dict__.pop(key, None)
		super().__setattr__(key, value)

	def __delattr__(self, item):
		"""
		Delete cached settings.
		"""
		super().__delattr__(item)
		self.__dict__.pop(item, None)

	@property
	def configured(self):
		"""
		Return True if the settings have already been configured.
		"""
		return self._settings is not None

	def reset(self):
		"""
		Clear the current wrapped settings. Useful when reloaded or in tests.
		"""
		self._settings = None


settings = LazySettings()



"""
This file has been forked from Django and PyDispatcher.
The PyDispatcher is licensed under BSD.
"""
import threading
import weakref
import logging
import asyncio

from pyplanet.core.exceptions import SignalException, SignalGlueStop
from pyplanet.utils.log import handle_exception


def _make_id(target):
	if hasattr(target, '__func__'):
		return id(target.__self__), id(target.__func__)
	return id(target)


NONE_ID = _make_id(None)
NO_RECEIVERS = object()

logger = logging.getLogger(__name__)


class Signal:
	"""
	A signal is a destination tho distribute to where multiple listeners get the message. (event distribution).
	"""

	def __init__(self, code=None, namespace=None, process_target=None, use_caching=False):
		"""
		Create a new signal.
		"""
		if not process_target:
			process_target = self.process
		self.process_target = process_target

		self.receivers = list()
		self.self_refs = dict()
		self.lock = threading.Lock()

		if code:
			self.code = code
		else:
			self.code = self.Meta.code

		if namespace:
			self.namespace = namespace
		else:
			self.namespace = self.Meta.namespace

		self.use_caching = use_caching
		self.sender_receivers_cache = weakref.WeakKeyDictionary() if use_caching else {}
		self._dead_receivers = False

	class Meta:
		"""
		The meta-class contains the code of the signal, used for string notation.
		An optional namespace could be given to override the app label namespace.

		.. warning::

			Only change or access this if you override the ``Signal`` class in your own class.

		"""
		code = None
		namespace = None

	async def process(self, **data):
		"""
		This method processed data into abstract data. You can give your own function in the init of the Signal or
		override the method.

		:param data: Raw data input
		:return: Parsed data output
		"""
		return data

	def has_listeners(self):
		"""
		Has the signal listeners.

		:return:
		"""
		return bool(self._live_receivers())

	def set_self(self, receiver, slf):  # pragma: no cover
		"""
		Set the self instance on a receiver.

		.. deprecated:: 0.0.1

		:param receiver: Receiver function.
		:param slf: Self instance
		"""
		with self.lock:
			lookup_key = _make_id(receiver)
			for key, _ in self.receivers:
				if lookup_key == key:
					ref = weakref.ref
					slf = ref(slf)
					self.self_refs[lookup_key] = slf
					return
			raise Exception('Receiver is not yet known! You registered too early!')

	def register(self, receiver, weak=True, dispatch_uid=None):
		"""
		Connect receiver to sender for signal.

		:param receiver: A function or an instance method which is to receive signals. Receivers must be hashable objects.
			If weak is True, then receiver must be weak referenceable.Receivers must be able to accept keyword arguments.
			If a receiver is connected with a dispatch_uid argument, it
			will not be added if another receiver was already connected with that dispatch_uid.

		:param weak: Whether to use weak references to the receiver. By default, the
			module will attempt to use weak references to the receiver
			objects. If this parameter is false, then strong references will
			be used.

		:param dispatch_uid: An identifier used to uniquely identify a particular instance of
			a receiver. This will usually be a string, though it may be anything hashable.
		"""
		if dispatch_uid:
			lookup_key = dispatch_uid
		else:
			lookup_key = _make_id(receiver)

		if weak:
			ref = weakref.ref
			receiver_object = receiver

			# Check for bound methods.
			if hasattr(receiver, '__self__') and hasattr(receiver, '__func__'):
				ref = weakref.WeakMethod
				receiver_object = receiver.__self__
			receiver = ref(receiver)
			weakref.finalize(receiver_object, self._remove_receiver)

		with self.lock:
			self._clear_dead_receivers()
			for rec_key in self.receivers:
				if rec_key == lookup_key:
					break
			else:
				self.receivers.append((lookup_key, receiver))
			self.sender_receivers_cache.clear()

	def unregister(self, receiver=None, dispatch_uid=None):
		"""
		Disconnect receiver from sender for signal.
		If weak references are used, disconnect need not be called. The receiver
		will be removed from dispatch automatically.

		:param receiver: The registered receiver to disconnect. May be none if dispatch_uid is specified.
		:param dispatch_uid: the unique identifier of the receiver to disconnect
		"""
		if dispatch_uid:
			lookup_key = dispatch_uid
		else:
			lookup_key = _make_id(receiver)

		disconnected = False
		with self.lock:
			self._clear_dead_receivers()
			for index in range(len(self.receivers)):
				(rec_key, _) = self.receivers[index]
				if rec_key == lookup_key:
					disconnected = True
					del self.receivers[index]
					if rec_key in self.self_refs:
						del self.self_refs[rec_key]
					break
			self.sender_receivers_cache.clear()

		return disconnected

	@staticmethod
	async def execute_receiver(receiver, args, kwargs, ignore_exceptions=False):
		try:
			if asyncio.iscoroutinefunction(receiver):
				if len(args) > 0:
					return receiver, await receiver(*args, **kwargs)
				return receiver, await receiver(**kwargs)

			if len(args) > 0:
				return receiver, receiver(*args, **kwargs)
			return receiver, receiver(**kwargs)
		except Exception as exc:
			if not ignore_exceptions:
				raise

			logger.exception(SignalException(
				'Signal receiver \'{}\' => {} thrown an exception!'.format(receiver.__module__, receiver.__name__)
			), exc_info=False)

			# Handle, will send to sentry if it's related to the core/contrib apps.
			handle_exception(exc, receiver.__module__, receiver.__name__)

			# Log the actual exception.
			logger.exception(exc)
			return receiver, exc

	async def send(self, source, raw=False, catch_exceptions=False, gather=True):
		"""
		Send signal with source.
		If any receiver raises an error, the error propagates back through send,
		terminating the dispatch loop. So it's possible that all receivers
		won't be called if an error is raised.

		:param source: The data to be send to the processor which produces data that will be send to the receivers.
		:param raw: Optional bool parameter to just send the source to the receivers without any processing.
		:param catch_exceptions: Catch and return the exceptions.
		:param gather: Execute multiple receivers at the same time (parallel). On by default!

		:return: Return a list of tuple pairs [(receiver, response), ... ].
		"""
		if raw is False:
			try:
				kwargs = await self.process_target(signal=self, source=source)
			except SignalGlueStop:
				# Stop calling the receivers when our glue says we should!
				return []
		else:
			kwargs = dict(**source, signal=self)

		if not self.receivers:
			return []

		# Prepare the responses from the calls.
		responses = []
		gather_list = []
		for key, receiver in self._live_receivers():
			# Dereference the weak reference.
			slf = self.self_refs.get(key, None)
			if slf and isinstance(slf, weakref.ReferenceType):
				slf = slf()
			args = [slf] if slf else []

			# Execute the receiver.
			coro = self.execute_receiver(receiver, args, kwargs, ignore_exceptions=catch_exceptions)
			if gather:
				gather_list.append(coro)
			else:
				responses.append(await coro)

		# If gather, wait on the asyncio.gather operation and return the responses from there.
		if gather:
			return await asyncio.gather(*gather_list)

		# Done, respond with all the results
		return responses

	async def send_robust(self, source=None, raw=False, gather=True):
		"""
		Send signal from sender to all connected receivers catching errors.

		:param source: The data to be send to the processor which produces data that will be send to the receivers.
		:param raw: Optional bool parameter to just send the source to the receivers without any processing.
		:param gather: Execute multiple receivers at the same time (parallel). On by default!

		:return: Return a list of tuple pairs [(receiver, response), ... ].
			If any receiver raises an error (specifically any subclass of Exception),
			return the error instance as the result for that receiver.
		"""
		return await self.send(source, raw, catch_exceptions=True, gather=gather)

	def _clear_dead_receivers(self):
		if self._dead_receivers:
			self._dead_receivers = False
			new_receivers = []
			for rec in self.receivers:
				if isinstance(rec[1], weakref.ReferenceType) and rec[1]() is None:
					continue
				new_receivers.append(rec)
			self.receivers = new_receivers

	def _live_receivers(self):
		"""
		Filter sequence of receivers to get resolved, live receivers.
		This checks for weak references and resolves them, then returning only
		live receivers.
		"""
		# We don't use the sender. Set it to none.
		sender = None

		receivers = None
		if self.use_caching and not self._dead_receivers:
			receivers = self.sender_receivers_cache.get(sender)
			# We could end up here with NO_RECEIVERS even if we do check this case in
			# .send() prior to calling _live_receivers() due to concurrent .send() call.
			if receivers is NO_RECEIVERS:
				return []

		if receivers is None:
			with self.lock:
				self._clear_dead_receivers()
				receivers = []
				for receiverkey, receiver in self.receivers:
					receivers.append((receiverkey, receiver))
				if self.use_caching:
					if not receivers:
						self.sender_receivers_cache[sender] = NO_RECEIVERS
					else:
						# Note, we must cache the weakref versions.
						self.sender_receivers_cache[sender] = receivers
		non_weak_receivers = []

		for receiver in receivers:
			key = receiver[0]
			receiver = receiver[1]

			if isinstance(receiver, weakref.ReferenceType):
				# Dereference the weak reference.
				receiver = receiver()
				if receiver is not None:
					non_weak_receivers.append((key, receiver))
			else:
				non_weak_receivers.append((key, receiver))
		return non_weak_receivers

	def _remove_receiver(self):
		# The list must be marked as dead. And will be cleaned in the next registry or call.
		# We can't directly remove because GC is always running when lock is preserved.
		self._dead_receivers = True



"""
The database registry class is used for managing apps models and states related to the Peewee backed ORM.
"""
import os
import inspect
import importlib

from glob import glob


def get_app_models(app):
	from .model import Model
	try:
		models_module = importlib.import_module('{}.models'.format(app.name))
		for name, obj in inspect.getmembers(models_module):
			if inspect.isclass(obj) and issubclass(obj, Model):
				if obj.__name__ == 'TimedModel':
					continue
				yield name, obj
	except Exception as e:
		return list()


def get_app_migrations(app):
	try:
		migration_files = glob(os.path.join(app.path, 'migrations', '*.py'))
		migration_files.sort()

		for file in migration_files:
			dir, file_name = os.path.split(file)
			name, ext = file_name.split('.')
			yield file, dir, name, ext
	except:
		return list()


class Registry:
	def __init__(self, instance, db):
		self.instance = instance
		self.db = db

		# Cache app contexts
		self.app_configs = dict()
		self.app_models = dict()
		self.app_migrations = dict()

		# Models are saved as [app_label.model_classname] = (app, name, model)
		self.models = dict()

	def init_app(self, app):
		"""
		Initiate the app database assets.
		:param app: App instance
		:type app: pyplanet.apps.AppConfig
		"""
		# Determinate correct paths.
		root_path = app.path
		models_path = os.path.join(root_path, 'models') \
			if os.path.exists(os.path.join(root_path, 'models')) and os.path.isdir(os.path.join(root_path, 'models')) \
			else root_path
		migrations_path = os.path.join(root_path, 'migrations')

		# Set paths in the config context.
		if os.path.exists(models_path):
			if not os.path.exists(migrations_path):
				try:
					os.mkdir(migrations_path)
				except:
					pass

		# Import the model module.
		self.app_models[app.label] = models = list(get_app_models(app))
		for name, model in models:
			self.models['{}.{}'.format(app.label, name)] = app, name, model

		# Import migration files.
		self.app_migrations[app.label] = migrations = list(get_app_migrations(app))



import asyncio

from pyplanet.views import TemplateView
from pyplanet.views.generics.list import ManualListView


class CommandsListView(ManualListView):
	title = 'Available chat commands'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Puzzle'

	is_admin_view = False

	data = []

	def __init__(self, app, player, is_admin_view):
		super().__init__(self)
		self.app = app
		self.instance = app.instance
		self.manager = app.context.ui
		self.player = player
		self.is_admin_view = is_admin_view

		self.child = None

	async def get_fields(self):
		return [
			{
				'name': 'Command',
				'index': 'command',
				'sorting': True,
				'searching': True,
				'width': 70,
				'type': 'label',
				'action': self.open_command_details
			},
			{
				'name': 'Description',
				'index': 'description',
				'sorting': False,
				'searching': True,
				'width': 150
			},
		]

	async def get_data(self):
		data = []

		self.title = 'Available admin chat commands' if self.is_admin_view else 'Available chat commands'

		for command in [c for c in self.instance.command_manager._commands if c.admin is self.is_admin_view]:
			if not self.is_admin_view or await command.has_permission(self.instance, self.player):
				command_text = ''
				if command.namespace:
					command_text += '|'.join(command.namespace) if isinstance(command.namespace, (list, tuple)) else command.namespace
					command_text += ' '
				command_text += command.command
				data.append({'command': command_text, 'description': command.description, 'command_object': command})

		data.sort(key=lambda c: c['command'])
		return data

	async def display(self, **kwargs):
		kwargs['player'] = self.player
		return await super().display(**kwargs)

	async def open_command_details(self, player, values, row, **kwargs):
		if self.child:
			return

		# Show details view.
		self.child = CommandDetailsView(self, self.player, row['command_object'])
		await self.child.display()
		await self.child.wait_for_response()
		await self.child.destroy()
		await self.display()  # refresh.
		self.child = None


class CommandDetailsView(TemplateView):
	template_name = 'core.pyplanet/command/details.xml'

	def __init__(self, parent, player, command):
		"""
		Initiate child edit view.

		:param parent: Parent view.
		:param player: Player instance.
		:param command: Command instance.
		:type parent: pyplanet.view.base.View
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type command: pyplanet.contrib.command.Command
		"""
		super().__init__(parent.manager)
		self.parent = parent
		self.player = player
		self.command = command

		self.response_future = asyncio.Future()

		self.subscribe('button_close', self.close)

	async def display(self, **kwargs):
		await super().display(player_logins=[self.player.login])

	async def get_context_data(self):
		context = await super().get_context_data()
		context['command_chat'] = str(self.command)
		context['command'] = self.command

		return context

	async def wait_for_response(self):
		return await self.response_future

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]

		await self.hide(player_logins=[player.login])
		self.response_future.set_result(None)
		self.response_future.done()



import os
import signal
import threading
import logging
import sys
import atexit

try:
	import fcntl
except:
	fcntl = dict()  # Windows only.

from pyplanet.conf import settings
from pyplanet.core.management import BaseCommand
from pyplanet.god.pool import EnvironmentPool
from pyplanet.utils.log import initiate_logger


class Command(BaseCommand):  # pragma: no cover
	help = 'Start the PyPlanet god with all it\'s subprocesses.'

	requires_system_checks = False
	requires_migrations_checks = False

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.detached = False
		self.pid_file = None
		self.pid = None

		self.pool = None

	def add_arguments(self, parser):
		parser.add_argument('--max-restarts', type=int, default=0)
		parser.add_argument('--tokio', dest='tokio', action='store_true')
		parser.add_argument('--detach', dest='detach', action='store_true')
		parser.add_argument('--pid-file', type=str, default=None)

	def handle(self, *args, **options):
		# Detach when asked.
		if 'detach' in options and options['detach']:
			self.detach()

		# Write process ID (pid) to file.
		self.write_pid(pid_file=options['pid_file'] if 'pid_file' in options and options['pid_file'] else 'pyplanet.pid')

		# Reload settings with error messages.
		settings._setup()

		# Initiate the logger.
		threading.current_thread().setName('Main')
		initiate_logger()
		logger = logging.getLogger(__name__)

		# Initiate the settings by accessing one.
		logger.debug('Initiated configuration and environment... (debug on, means no error reporting and verbose output')
		if not settings.DEBUG:
			logger.info('Initiated configuration and environment...')
		logger.info('-------------------------------[  PyPlanet v{}  ]-------------------------------'.format(self.get_version()))

		# Start god process (the current started process).
		self.pool = EnvironmentPool(settings.POOLS, max_restarts=options['max_restarts'], options=options)
		self.pool.populate()

		# Starting all processes.
		self.pool.start()

		# Start the watchdog.
		try:
			self.pool.watchdog()
		except KeyboardInterrupt:
			self.pool.shutdown()
			exit(0)

	def write_pid(self, pid_file):
		self.pid_file = pid_file

		# Check for old pid file.
		old_pid = ''
		if os.path.isfile(pid_file):
			with open(pid_file, 'r') as pid_handle:
				old_pid = pid_handle.read()

		# Create lock file.
		try:
			lock_file = open(pid_file, 'w')
		except IOError:
			logging.error('Can\'t create PID file!')
			sys.exit(1)

		# Try to get lock on the file (on non-Windows).
		if fcntl != dict():
			try:
				fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
			except IOError:
				logging.error('Can\'t create PID file! Can\'t lock on file!')

				# We need to overwrite the pidfile if we got here.
				with open(pid_file, 'w') as pid_handle:
					pid_handle.write(old_pid)

		# Write PID file.
		lock_file.write(str(os.getpid()))
		lock_file.flush()

	def detach(self):
		# Fork to background.
		try:
			pid = os.fork()
		except OSError as e:
			logging.error('Unable to fork, error: {} ({})'.format(str(e), e.errno))
			sys.exit(1)

		if pid != 0:
			# This is the parent process. Exit.
			os._exit(0)

		# Stop listening for signals.
		os.setsid()
		self.pid = os.getpid()

		# Determinate /dev/null on machine.
		dev_null = '/dev/null'
		if hasattr(os, 'devnull'):
			dev_null = os.devnull

		# Redirect output to dev null.
		dev_null_fd = os.open(dev_null, os.O_RDWR)
		os.dup2(dev_null_fd, 0)
		os.dup2(dev_null_fd, 1)
		os.dup2(dev_null_fd, 2)
		os.close(dev_null_fd)

		self.detached = True

		# Register shutdown methods.
		signal.signal(signal.SIGTERM, self.sigterm)
		signal.signal(signal.SIGHUP, self.sigterm)
		atexit.register(self.exit)

	def sigterm(self, signum, frame):
		self.pool.shutdown()
		sys.exit(0)

	def exit(self):
		os.remove(self.pid_file)



"""
Exceptions for Setting Manager.
"""


class SettingException(Exception):
	"""Abstract setting exception."""


class SerializationException(SettingException):
	"""Setting value (un)serialization problems"""


class TypeUnknownException(SettingException):
	"""The type is unknown."""





class ParamException(Exception):
	pass


class ParamParseException(ParamException):
	pass


class ParamValidateException(ParamException):
	pass


class NotValidated(Exception):
	"""
	Your parser hasn't been called with .parse() before, so no parsing took place!
	"""
	pass


class InvalidParamException(Exception):
	"""
	Invalid parameter arguments given!
	"""



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models.karma import Karma


def upgrade(migrator: SchemaMigrator):
	expanded_score_field = FloatField(null=True)
	migrate(
		migrator.drop_not_null(Karma._meta.db_table, 'score'),
		migrator.add_column(Karma._meta.db_table, 'expanded_score', expanded_score_field),
	)


def downgrade(migrator: SchemaMigrator):
	pass



import asyncio
import logging

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.core import Controller
from pyplanet.core.ui.template import load_template
from pyplanet.utils.codeutils import deprecated
from pyplanet.views.base import View


class TemplateView(View):
	"""
	The TemplateView will provide a view based on a XML template (ManiaLink for example).
	The view contains some class properties that are required to work. Those are described bellow.

	To use the TemplateView. Initiate it in your own View class, and override one of the following methods:

	:method get_context_data(): Return the global context data here.
								Make sure you use the super() to retrieve the current context.
	:method get_all_player_data(logins): Retrieve the player specific dictionary.
										Return dict with player as key and value should contain the data dict.
	:method get_per_player_data(login): Retrieve the player specific dictionary per player.
										Return dict with the data dict for the specific login (player).
	:method get_template(): Return the template instance from Jinja2. You mostly should not override this method.

	As alternative you can manipulate the instance.data and instance.player_data too.

	**Properties that are useful to change**:

	:prop data: Global context data. Dict.
	:prop player_data: Player context data. Dict with player as key.
	:prop hide_click: Should the manialink disappear after clicking a button/text.
	:prop timeout: Timeout to hide manialink in seconds.

	**Example usage:**

	.. code-block:: python

		class AlertView(TemplateView):
			template_name = 'my_app/test.xml' # template should be in: ./my_app/templates/test.xml
			# Some prefixes that can be used in the template_name:
			#
			# - core.views: ``pyplanet.views.templates``.
			# - core.pyplanet: ``pyplanet.apps.core.pyplanet.templates``.
			# - core.maniaplanet: ``pyplanet.apps.core.pyplanet.templates``.
			# - core.trackmania: ``pyplanet.apps.core.trackmania.templates``.
			# - core.shootmania: ``pyplanet.apps.core.shootmania.templates``.
			# - [app_label]: ``[app path]/templates``.

			async def get_context_data(self):
				context = await super().get_context_data()
				context['title'] = 'Sample'
				return context

	"""

	template_name = None

	async def get_context_data(self):
		"""
		Get global and local context data, used to render template.
		"""
		context = dict(
			id=self.id,
		)
		return context

	@deprecated
	async def get_player_data(self):
		"""
		Get data per player, return dict with login => data dict.

		.. deprecated:: 0.4.0
			Use :func:`get_per_player_data` and :func:`get_all_player_data` instead. Will be removed in 0.8.0!

		"""
		return dict()

	async def get_all_player_data(self, logins):
		"""
		Get all player data, should return dictionary with login as key, and dict as value.

		:param logins: Login list of players. String list.
		:return: Dictionary with data.
		"""
		return dict()

	async def get_per_player_data(self, login):
		"""
		Get data for specific player. Will be called for all players that will render the xml for.

		:param login: Player login string.
		:return: Dictionary or None to ignore.
		:type login: str
		"""
		return dict()

	async def get_template(self):
		return await load_template(self.template_name)

	async def render(self, *args, player_login=None, **kwargs):
		"""
		Render template for player. This will only render the body and return it. Not send it!

		:param player_login: Render data only for player, set to None to globally render (and ignore player_data).
		:return: Body, rendered manialink + script.
		"""
		kwargs['data'] = await self.get_context_data()
		kwargs['player_login'] = player_login
		kwargs['player_data'] = self.player_data # Should already been read by display().
		kwargs['template'] = await self.get_template()
		return await super().render(*args, **kwargs)

	async def display(self, player_logins=None, **kwargs):
		"""
		Display the manialink. Will also render if no body is given. Will show per player or global. depending on
		the data given and stored!

		:param player_logins: Only display to the list of player logins given.
		"""
		# Get player data (old way). Deprecated since 0.4.0.
		# Added warning since 0.8.0.
		self.player_data = deprecated_data = await self.get_player_data()
		# COMMENTED IN 0.8.2: FAST FIX.
		# if deprecated_data and len(deprecated_data.keys()) > 0:
			# logging.getLogger(self.__class__.__module__ + '.' + self.__class__.__name__).warning(
			# 	'Method \'get_player_data\' is deprecated and will be removed soon! Please refactor your code!'
			# )
			# self.player_data = deprecated_data

		self.player_data.update(await self.get_all_player_data(
			player_logins or [p.login for p in Controller.instance.player_manager.online]
		))

		# Get player data (new way).
		async def get_player_data(login):
			data = await self.get_per_player_data(login)
			if not data or isinstance(data, dict) and len(data.keys()) == 0:
				data = None
			return login, data

		player_data = await asyncio.gather(*[
			get_player_data(p.login) if isinstance(p, Player) else get_player_data(p)
			for p in player_logins or Controller.instance.player_manager.online
		])

		# TODO: This can be flatten with `self.player_data = dict(player_data)` after deprecated code has been removed.
		for login, data in player_data:
			if not isinstance(data, dict) or len(data.keys()) == 0:
				continue
			if login in self.player_data and isinstance(self.player_data[login], dict):
				self.player_data[login].update(data)
			else:
				self.player_data[login] = data

		# Fallback in case the data is empty or not a dictionary.
		if not isinstance(self.player_data, dict):
			self.player_data = dict()

		return await super().display(player_logins, **kwargs)



from peewee import *

from pyplanet.core.db import Model


class Permission(Model):
	namespace = CharField(
		max_length=150,
		null=False,
		help_text='Namespace of the permission. Mostly the app.label.'
	)

	name = CharField(
		max_length=150,
		null=False,
		help_text='Name of permission, in format {app_name|core}:{name}'
	)

	description = TextField(
		null=True, default=None, help_text='Description of permission.'
	)

	min_level = IntegerField(
		default=1, help_text='Minimum required player level to be able to use this permission.'
	)

	class Meta:
		indexes = (
			(('namespace', 'name'), True),
		)





class Zone:
	def __init__(self, path, continent, country=None, province=None):
		self.path = path
		self.continent = continent
		self.country = country
		self.province = province


def parse_path(path: str) -> Zone:
	parts = path.split('|')

	continent = None
	country = None
	province = None

	if len(parts) > 1:
		continent = parts[1]
	if len(parts) > 2:
		country = parts[2]
	if len(parts) > 3:
		province = parts[3]

	return Zone(path, continent, country, province)



import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

from . import views


class Info(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.map_widget = None
		self.server_widget = None

		self.update_pending = False

	async def on_start(self):
		self.context.signals.listen(mp_signals.map.map_begin, self.map_begin)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.player.player_disconnect, self.any_change)
		self.context.signals.listen(mp_signals.player.player_info_changed, self.any_change)

		# Move the multilapinfo a bit. (Only Trackmania).
		if self.instance.game.game == 'tm':
			self.instance.ui_manager.properties.set_attribute('multilap_info', 'pos', '107., 88., 5.')

		if self.instance.game.game in ['tm', 'sm']:
			self.instance.ui_manager.properties.set_visibility('map_info', False)

		self.map_widget = views.MapInfoWidget(self)
		self.server_widget = views.ServerInfoWidget(self)

		# Don't wait on the displaying of the widget.
		asyncio.ensure_future(self.map_widget.display())
		asyncio.ensure_future(self.server_widget.display())

		# Update server widget.
		asyncio.ensure_future(self.server_update_loop())

	async def server_update_loop(self):
		while True:
			await asyncio.sleep(5)
			try:
				if self.update_pending:
					await self.server_widget.display()
					self.update_pending = False
			except:
				pass

	async def map_begin(self, map):
		await asyncio.gather(
			self.map_widget.display(),
			self.server_widget.display()
		)

	async def player_connect(self, player, is_spectator, source, signal):
		await asyncio.gather(
			self.map_widget.display(player=player),
			self.server_widget.display(player=player)
		)
		self.update_pending = True

	async def any_change(self, **kwargs):
		self.update_pending = True



import logging
import os
from argparse import Namespace

from pyplanet.views.generics import ManualListView
from pyplanet.utils import gbxparser

logger = logging.getLogger(__name__)


class BrowserView(ManualListView):
	app = None
	title = 'Local Maps'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Browse'

	data = []

	def __init__(self, app, player):
		super().__init__(self)
		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.objects_raw = []
		self.current_dir = ''

		self.fields = self.create_fields()

		self.sort_field = self.fields[0]

	async def set_dir(self, directory):
		self.objects_raw = list()
		self.current_dir = directory
		self.title = 'Local Maps     $n$aaa' + self.current_dir

		full_dir = os.path.join('UserData', 'Maps', self.current_dir)
		files = await self.app.instance.storage.driver.listdir(path=full_dir)

		self.objects_raw.append({'icon': '$ff0D', 'file_name': '..'})

		# Add the files to the directory list.
		for file in files:
			is_file = await self.app.instance.storage.driver.is_file(os.path.join(full_dir, file))
			if is_file and not file.lower().endswith('map.gbx'):
				continue
			self.objects_raw.append({'icon': 'F' if is_file else '$ff0D', 'file_name': file})
		await self.display(player=self.player.login)

	def create_fields(self):
		return [
			{
				'name': 'T',
				'index': 'icon',
				'sorting': True,
				'searching': False,
				'width': 10,
				'type': 'label'
			},
			{
				'name': 'Name',
				'index': 'file_name',
				'sorting': True,
				'searching': True,
				'width': 150,
				'type': 'label',
				'action': self.action_file
			},
		]

	async def get_fields(self):
		return self.fields

	async def action_file(self, player, values, instance, **kwargs):
		isdir = instance['icon'] == '$ff0D'
		filename = instance['file_name']
		if isdir:
			if filename == '..':
				await self.set_dir(os.path.dirname(self.current_dir))
			else:
				await self.set_dir(os.path.join(self.current_dir, filename))
		else:
			await self.add_map(filename, player)

	async def add_map(self, filename, player):
		map_path = os.path.join(self.current_dir, filename)

		data = Namespace()
		data.map = map_path
		await self.app.map.add_local_map(player, data)



class Vote:
	"""
	Vote object, keeping track of all votes.
	"""

	def __init__(self):
		"""
		Initializes the vote.
		"""
		self.action = ''
		self.requester = None
		self.time_limit = None
		self.votes_required = 0
		self.votes_current = []

		self.vote_added = None
		self.vote_removed = None
		self.vote_passed = None

	async def add_vote(self, player):
		"""
		Called to add a vote by a player to the object.

		:param player: player who voted for the vote
		"""

		if player.login in self.votes_current:
			return False

		self.votes_current.append(player.login)

		await self.fire_added_event(player)
		if len(self.votes_current) >= self.votes_required:
			await self.fire_passed_event()

		return True

	async def remove_vote(self, player):
		"""
		Called to remove a vote by a player to the object.

		:param player: player who against the vote
		"""

		if player.login in self.votes_current:
			self.votes_current.remove(player.login)
			await self.fire_removed_event(player)
			return True

		return False

	async def fire_added_event(self, player):
		"""
		Calls the event for when a vote has been added.

		:param player: player who voted for the vote
		"""

		if self.vote_added is not None:
			await self.vote_added(vote=self, player=player)

	async def fire_removed_event(self, player):
		"""
		Calls the event for when a vote has been removed.

		:param player: player who against the vote
		"""

		if self.vote_removed is not None:
			await self.vote_removed(vote=self, player=player)

	async def fire_passed_event(self, forced=False):
		"""
		Calls the event for when the vote has passed.

		:param forced: whether the vote was forced by the admin
		"""

		if self.vote_passed is not None:
			await self.vote_passed(vote=self, forced=forced)



import asyncio

from tests.integration import ControllerTestCase

#
# class TestPlayerManager(ControllerTestCase):
#
# 	async def test_connect(self):
# 		before_len = len(self.instance.player_manager.online)
#
# 		await self.instance.gbx.execute('ConnectFakePlayer')
# 		await asyncio.sleep(1)
#
# 		after_len = len(self.instance.player_manager.online)
#
# 		assert before_len != after_len
#
# 	async def test_disconnect(self):
# 		before_len = len(self.instance.player_manager.online)
#
# 		name = await self.instance.gbx.execute('ConnectFakePlayer')
# 		await asyncio.sleep(1)
#
# 		after_len = len(self.instance.player_manager.online)
#
# 		await self.instance.gbx.execute('DisconnectFakePlayer', name)
# 		await asyncio.sleep(1)
#
# 		discon_len = len(self.instance.player_manager.online)
#
# 		assert before_len != after_len



from pyplanet.core import Controller
from pyplanet.core.events import Callback, handle_generic


async def handle_royal_points(source, signal, **kwargs):
	player = await Controller.instance.player_manager.get_player(login=source['login'])
	return dict(player=player, type=source['type'], points=source['points'])


player_score_points = Callback(
	call='Script.Shootmania.Royal.Points',
	namespace='shootmania',
	code='royal_player_score_points',
	target=handle_royal_points
)
"""
:Signal: 
	Player score points.
:Code:
	``shootmania:royal_player_score_points``
:Description:
	Callback sent when a player scores some points.
:Original Callback:
	`Script` Shootmania.Royal.Points

:param player: Player instance.
:param type: Type of score, like 'Pole', 'Hit', or 'Survival'.
:param points: Points that the player gains.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


player_spawn = Callback(
	call='Script.Shootmania.Royal.PlayerSpawn',
	namespace='shootmania',
	code='royal_player_spawn',
	target=handle_generic
)
"""
:Signal: 
	Player spawns.
:Code:
	``shootmania:royal_player_spawn``
:Description:
	Callback sent when a player is spawned.
:Original Callback:
	`Script` Shootmania.Royal.PlayerSpawn

:param player: Player instance.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""


results = Callback(
	call='Script.Shootmania.Royal.RoundWinner',
	namespace='shootmania',
	code='royal_results',
	target=handle_generic
)
"""
:Signal: 
	End of round with the winner of the Royal round.
:Code:
	``shootmania:royal_results``
:Description:
	Callback sent at the end of the round with the player instance of the winner.
:Original Callback:
	`Script` Shootmania.Royal.RoundWinner

:param player: Player instance that won the round.
:type player: pyplanet.apps.core.maniaplanet.models.player.Player
"""



from .messages import (
	CRITICAL, DEBUG, ERROR, INFO, WARNING,
	CheckMessage, Critical, Debug, Error, Info, Warning,
)
from .registry import run_checks, register, tag_exists

__all__ = [
	'CRITICAL', 'DEBUG', 'ERROR', 'INFO', 'WARNING',
	'run_checks', 'register', 'tag_exists',
	'CheckMessage', 'Critical', 'Debug', 'Error', 'Info', 'Warning',
]



"""
Sector Times.
"""
import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.sector_times.views import SectorTimesWidget, CheckpointDiffWidget, GearIndicatorView
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.setting import Setting
from pyplanet.core.signals import pyplanet_start_after


class SectorTimes(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.sector_widget = SectorTimesWidget(self)
		self.cp_widget = CheckpointDiffWidget(self)
		self.gear_view = GearIndicatorView(self)

		self.setting_enable_gear_indicator = Setting(
			'gear_indicator', 'Enable Gear Indicator', Setting.CAT_FEATURES, type=bool, default=True,
			description='Enable the Gear Indicator View in Stadium based titles',
			change_target=self.reload_settings
		)
		self.gear_view_possible = False

	async def on_start(self):
		await self.context.setting.register(
			self.setting_enable_gear_indicator
		)

		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.map.map_start, self.map_start)
		self.context.signals.listen(mp_signals.flow.podium_start, self.podium_start)

		self.context.signals.listen(pyplanet_start_after, self.on_after_start)

		self.gear_view_possible = \
			self.instance.map_manager.current_map.environment == 'Stadium' and self.instance.game.game == 'tm'

		# Set the MP CP diff widget position.
		if self.instance.game.game in ['tm', 'sm']:
			self.instance.ui_manager.properties.set_attribute('checkpoint_time', 'pos', '0. 8. -10.')

	def is_mode_supported(self, mode):
		return mode != 'Trackmania/TM_RoyalTimeAttack_Online'

	async def on_after_start(self, *args, **kwargs):
		await asyncio.sleep(1)
		asyncio.ensure_future(asyncio.gather(*[
			self.player_connect(p) for p in self.instance.player_manager.online
		]))

	async def player_connect(self, player, **kwargs):
		await self.sector_widget.display(player)
		await self.cp_widget.display(player)
		if self.gear_view_possible and await self.setting_enable_gear_indicator.get_value():
			await self.gear_view.display(player.login)

	async def map_start(self, *args, **kwargs):
		await asyncio.sleep(2)
		await self.sector_widget.display()
		await self.cp_widget.display()

	async def podium_start(self, *args, **kwargs):
		await self.sector_widget.hide()
		await self.cp_widget.hide()

	async def reload_settings(self, *args, **kwargs):
		if await self.setting_enable_gear_indicator.get_value() and self.gear_view_possible:
			await self.gear_view.display()
		else:
			await self.gear_view.hide()




from pyplanet.core.checks import Error


class CheckRegistry:
	def __init__(self):
		self.registered_checks = set()
		self.deployment_checks = set()

	def register(self, check=None, **kwargs):
		def inner(check):
			checks = self.deployment_checks if kwargs.get('deploy') else self.registered_checks
			checks.add(check)
			return check

		if callable(check):
			return inner(check)
		return inner

	def run_checks(self, app_configs=None, tags=None, include_deployment_checks=False, instance=None):
		errors = list()
		checks = self.get_checks(include_deployment_checks)

		try:
			from pyplanet.conf import settings
			_ = settings.DEBUG
		except:
			errors.append(
				Error(
					'Settings module not found, make sure you set PYPLANET_SETTINGS or the Cli option --settings!'
				)
			)

		for check in checks:
			new_errors = check(app_configs=app_configs, instance=instance)
			try:
				new_errors_iter = iter(new_errors)
			except TypeError:
				raise Exception(
					'The function {} did not return a list. All functions registered with the checks registry must return a list.'.format(check)
				)

			errors.extend(new_errors)

		return errors

	def tag_exists(self, tag, include_deployment_checks=False):
		# TODO: Extend the check registry class.
		return False

	def get_checks(self, include_deployment_checks=False):
		checks = list(self.registered_checks)
		if include_deployment_checks:
			checks.extend(self.deployment_checks)
		return checks


registry = CheckRegistry()
register = registry.register
run_checks = registry.run_checks
tag_exists = registry.tag_exists



from pyplanet.apps.config import AppConfig


class ShootmaniaConfig(AppConfig):
	name = 'pyplanet.apps.core.shootmania'
	core = True

	app_dependencies = [
		'core.maniaplanet'
	]

	game_dependencies = [
		'shootmania'
	]



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from ..models import TestModel

sample_field = CharField(default='unknown')


def upgrade(migrator: SchemaMigrator):
	migrate(
		migrator.add_column(TestModel._meta.db_table, 'sample', sample_field)
	)


def downgrade(migrator: SchemaMigrator):
	pass



import asyncio
import asynctest

from pyplanet.core import Controller


class ControllerTestCase(asynctest.TestCase):

	async def setUp(self):
		self.instance = Controller.prepare(name='default').instance
		await self.instance._start()
		asyncio.ensure_future(self.instance.gbx.listen())

	async def tearDown(self):
		await self.instance.gbx.disconnect()
		try:
			await self.instance.db.engine.disconnect()
		except:
			pass
		del self.instance
